#!/usr/bin/python
# -*- coding: UTF-8 -*-

# This programm runs on BReaBox - a device that can play audio books for blind, 
# computerafraid people
#
# put it in /etc/rc.local with the call:
# python BReaBox_main.py

# ########################### Changeable Constants ############################

strBooksFolder = "/media/STICK/"
#strBooksFolder = "/Users/stefanstechert/Desktop/BReaBoxDemo/"
strText2SpeechFolder = "/home/pi/allgemeine_sprachausgaben/"
strBookPrefix = "Buch_aktiv_"
strBookAppendix = ".txt"
strChapterPrefix = "Kapitel_aktiv_"
strChapterAppendix = ".txt"
strUSBStickPrepared = "Stick_ist_vorbereitet.txt"
strLanguagePattern = "--lang=de-DE"
numWaitAtInit = 2

# ########################### Imports #########################################

import serial
import time
import math
import subprocess
import glob
import sys

#import string

# ########################### Functions begin #################################

# ----------------------------------------------------------------------------
# plays a mp3; non-Blocking means: do not wait for completion
# ----------------------------------------------------------------------------
def play(strMp3FilePath,strPlayAction):
	global strText2SpeechFolder

	if strPlayAction == "non-blocking":
		subprocess.Popen(["mplayer","-slave","-input","file=/tmp/mplayer-control","-ao","alsa:device=hw=0.0", strMp3FilePath + ".mp3"])
		
	elif strPlayAction == "end":
		subprocess.check_output(["mplayer","-ao","alsa:device=hw=0.0",strMp3FilePath + ".mp3"])
	
	elif strPlayAction == "stop":
		try:
			#with open("/tmp/mplayer-control","w") as out:
			#	out.write("quit\n")
			strShell = subprocess.check_output(["pidof","mplayer"])
			subprocess.check_output(["kill","-9",strShell[:-1]])
		except:
			pass
		
	elif strPlayAction == "pause":
		try:
			with open("/tmp/mplayer-control","w") as out:
				out.write("pause\n")
		except:
			pass
		
	elif strPlayAction == "unpause":
		try:
			with open("/tmp/mplayer-control","w") as out:
				out.write("pause\nmute 1\nseek -10\nmute 0\n")
		except:
			pass


# ----------------------------------------------------------------------------
# Opens the serial interface
# ----------------------------------------------------------------------------
def serialOpen():
	global serialInterface
	serialInterface = serial.Serial("/dev/ttyAMA0", 9600, timeout=1)
	serialInterface.open()


# ----------------------------------------------------------------------------
# Adjusts the volume
# ----------------------------------------------------------------------------
def adjustVolume(strNewSoundValue):
	global numLastSoundValue
	# magic numbers
	# headphones: HPOUT1 Digital Volume (191 is max but FAAAAR TOO LOUD! SERIOUS EAR DAMAGE !)
	# amixer cset numid=159 50
	numHeadphoneSetValue = min((int(strNewSoundValue) / 10),100)
	strShell = subprocess.check_output(["amixer","cset","numid=159",str(numHeadphoneSetValue)])
	
	# speaker: Speaker Digital Volume (191 is max but TOO LOUD!)
	# amixer cset numid=162 50
	numSpeakerSetValue = min((int(strNewSoundValue) / 7),150)
	strShell = subprocess.check_output(["amixer","cset","numid=162",str(numSpeakerSetValue)])

	print ("Volume: " + strNewSoundValue + " Headphone: " + str(numHeadphoneSetValue) + " Speaker: " + str(numSpeakerSetValue))
	numLastSoundValue = int(strNewSoundValue)


# ----------------------------------------------------------------------------
# Switches output: Speaker or headphones
# ----------------------------------------------------------------------------
def switchOutput(strNewOutput):
	global strLastOutput
	global numLastSoundValue

	if strNewOutput == "Headphones":
		strShell = subprocess.check_output(["/home/pi/use_case_scripts/Reset_paths.sh"])
		strShell = subprocess.check_output(["/home/pi/use_case_scripts/Playback_to_Headset.sh"])
	else:
		strShell = subprocess.check_output(["/home/pi/use_case_scripts/Reset_paths.sh"])
		strShell = subprocess.check_output(["/home/pi/use_case_scripts/Playback_to_Speakers.sh"])
	
	print "output is now switched to", strNewOutput

	# after switching the outputs everything is too loud !
	adjustVolume(str(numLastSoundValue))
	strLastOutput = strNewOutput


# ----------------------------------------------------------------------------
# List all files or folders under a specific folder
# ----------------------------------------------------------------------------
def listFiles(strPath,strFilesOrFolders):
	strShell = subprocess.check_output(["ls","-p1",strPath])
	cObjects = strShell.split("\n")
	cFiles = []
	cFolders = []

	for strObject in cObjects:
		if len(strObject)>0:
			if strObject[len(strObject)-1]=="/":
				cFolders.append(strObject[:len(strObject)-1])
			else:
				cFiles.append(strObject)

	if strFilesOrFolders == "Files":
		return cFiles
	else:
		return cFolders


# ----------------------------------------------------------------------------
# Filter out the text between a start and an end expression
# ----------------------------------------------------------------------------
def filterOutText(strText,strStart,strEnd):
	bValid = True
	if not(len(strText) < (len(strStart) + len(strEnd))):
		if (strText.find(strStart) >= 0) and (strText.find(strEnd) >= 0):
			numStartPos = strText.find(strStart) + len(strStart)
			numEndPos = strText.find(strEnd)
			if numStartPos > numEndPos:
				bValid = False
		else:
			bValid = False
	else:
			bValid = False

	if bValid:
		return strText[numStartPos:numEndPos]
	else:
		return ""


# ----------------------------------------------------------------------------
# read serial message, check for plaus. return first action found
# ----------------------------------------------------------------------------
def readSerialMessageAndPlausibilize():
	global serialInterface
	global numLastSoundValue
	global strLastOutput

	numChecksumSum = 0
	bMessageCorrect = True

	strNextAction = ""
	strParam = ""
	strMessage = ""

	# 5 Buttons, Switch, 4 Analogdigits, Chksum, Carrige return, Line feed
	#              000000anlgC
	#strMessage = "11000010014"
	numMessageLength = (5 + 1 + 4 + 1 + 1 + 1)

	serialInterface.flushInput()

	# try to recieve messages until length is correct  -----------
	while not(len(strMessage) == numMessageLength):
		strMessage = serialInterface.readline()

	# all numbers ? ---------------
	if bMessageCorrect:
		# -2=Carrige return, Line feed
		for i in range(0,numMessageLength - 2):
			strNumbers = "0123456789"
			strLetter = strMessage[i]
			if not (strNumbers.find(strLetter) >= 0):
				bMessageCorrect = False

	# check checksum ----------------
	if bMessageCorrect:
		# Magic numbers
		for i in range(0,6):
			numChecksumSum = numChecksumSum + int(strMessage[i])

		numChecksumSum = numChecksumSum + int(strMessage[6:10])

		if not((numChecksumSum % 9) == (int(strMessage[10]))):
			bMessageCorrect = False

	# display: is correct ? ----------------
	# dbg
	strMessage = strMessage.replace("\r"," ").replace("\n"," ")
	if bMessageCorrect:
		pass
		#print "Received Message:", strMessage, "Correct."

	else:
		#print "Received Message:", strMessage, "Incorrect."
		pass
	# choose one action from all interfaces ----------------
	if bMessageCorrect:
		# Magic numbers
		if math.fabs(int(strMessage[6:10])-numLastSoundValue)>10:
			strNextAction = "Volume"
			strParam =      strMessage[6:10]

		elif (strMessage[5] == "0") and (strLastOutput != "Headphones"):
			strNextAction = "Output"
			strParam =      "Headphones"

		elif (strMessage[5] == "1") and (strLastOutput != "Speaker"):
			strNextAction = "Output"
			strParam =      "Speaker"

		elif strMessage[0] == "1":
			strNextAction = "Book"
			strParam =      "Previous"

		elif strMessage[1] == "1":
			strNextAction = "Book"
			strParam =      "Next"

		elif strMessage[2] == "1":
			strNextAction = "Chapter"
			strParam =      "Previous"

		elif strMessage[3] == "1":
			strNextAction = "Chapter"
			strParam =      "Next"

		elif strMessage[4] == "1":
			strNextAction = "Pause"

	cCmd = [strNextAction,strParam]
	return cCmd


# ----------------------------------------------------------------------------
# looks, if a certain file exists
# ----------------------------------------------------------------------------
def lookIfFileExists(strCompleteFilePath):
	strShell = "nothing found"
	try:
		strShell = subprocess.check_output(["ls",strCompleteFilePath])
	except:
		pass
	if strShell == "nothing found":
		return False
	else:
		return True


# ----------------------------------------------------------------------------
# finds the MP3 file with the highest number
# ----------------------------------------------------------------------------
def findLastMP3File(strCompleteFolderPath):
	cFiles = listFiles(strCompleteFolderPath,"Files")

	numAct = 0
	numMax = 0
	for strFile in cFiles:
		# Magic numbers
		if len(strFile) == 7:
			if (strFile[-4:] == ".mp3"):
				if int(strFile[:-4]) > numMax:
					numMax = int(strFile[:len(strFile)-4])
					strReturnValue = strFile[:len(strFile)-4]

	return strReturnValue


# ----------------------------------------------------------------------------
# sets a bookmark for the actual chapter or the actual book
# ----------------------------------------------------------------------------
def setBookmark(strValue,strPath,strPrefix,strAppendix):
	# rm old bookmark
	cPatternMatches = glob.glob(strPath + strPrefix + "*" + strAppendix)
	for strFile in cPatternMatches:
		subprocess.check_output(["rm",strFile])

	# set new bookmark
	subprocess.check_output(["touch",strPath + strPrefix + strValue + strAppendix])


# ----------------------------------------------------------------------------
# gets a bookmark for the actual chapter or the actual book if none set deflt.
# ----------------------------------------------------------------------------
def getBookmark(strDefaultValue,strPath,strPrefix,strAppendix):
	cPatternMatches = glob.glob(strPath + strPrefix + "*" + strAppendix)
	if len(cPatternMatches)>0:
		return filterOutText(cPatternMatches[0],strPath + strPrefix, strAppendix)
	else:
		subprocess.check_output(["touch",strPath + strPrefix + strDefaultValue + strAppendix])
		return strDefaultValue


# ----------------------------------------------------------------------------
# Determines the title of the next or the previous book (even at start or end)
# ----------------------------------------------------------------------------
def determineNextOrPreviousBook(strBooksFolder,strBook,strParam):
	cFolders = listFiles(strBooksFolder,"Folders")
	for numFolder in range(0,len(cFolders)):
		if strBook == cFolders[numFolder]:
			if strParam == "Next":
				if numFolder == len(cFolders) - 1:
					return cFolders[0]
				else:
					return cFolders[numFolder + 1]
			else: # "Previous"
				if numFolder == 0:
					return cFolders[len(cFolders) - 1]
				else:
					return cFolders[numFolder - 1]


# ----------------------------------------------------------------------------
# reads a given text but instead outputting to speaker - it saves as MP3
# ----------------------------------------------------------------------------
def readAsMP3(strText,DestinationPath,strLanguagePattern):
	# pico2wave --lang=de-DE --wave=test.wav "Der Ast ist grüner als draußen"
	# lame -V2 -a --vbr-new --resample 32 test.wav test.mp3
	subprocess.check_output(["pico2wave",strLanguagePattern,"--wave=temp.wav","\"" + strText + "\""])
	subprocess.check_output(["lame","-V2","-a","temp.wav",DestinationPath])
	subprocess.check_output(["rm","temp.wav"])


# ----------------------------------------------------------------------------
# checks as a workaround for the soundcard - a few seconds after X boot the
# .asoundrc is overwritten
# ----------------------------------------------------------------------------
def isSoundCardWrong():
	strShell = subprocess.check_output(["cat",".asoundrc"])
	if (strShell.find("card 1")>-1):
		print strShell
		return True
	else:
		return False


# ----------------------------------------------------------------------------
# this chooses the right soundcard again
# ----------------------------------------------------------------------------	
def switchToCorrectSoundCard():
	subprocess.check_output(["cp","asoundrc_modified",".asoundrc"])
	print "SOUNDCARD CORRECTED !"	


# ----------------------------------------------------------------------------
# Shuts the Raspi down
# ----------------------------------------------------------------------------
def Shutdown():
		serialInterface.close()
		play("","stop")
		play(strText2SpeechFolder + "Shutdown", "end")
		subprocess.Popen(["sudo","shutdown","-h","now"])
		time.sleep(60)


# ########################### Variables ######################################
# ------------- Globals (no statics in python :-( ) -------------

serialInterface = 0
numLastSoundValue = -1000
strLastOutput = "Init"

# Columns of cCmd (the resulting action from all Controls the box has) 
idxAction = 0
idxParam = 1

# ########################### Main Programm ##################################

strState = "Init"
while True:
	if strState == "Init": # --------------------------------------------
		print "Init---"
		serialOpen()
		
		# workaround: do not start with the programm, until sound card is corrected
		# therefore check, if programm was called with parameters or is in debug mode
		if len(sys.argv) < 2:
			bSoundCardIsCorrected = True
		else:
			bSoundCardIsCorrected = False
			
		while not(bSoundCardIsCorrected):
			if isSoundCardWrong():
				switchToCorrectSoundCard()
				bSoundCardIsCorrected = True
				time.sleep(0.1)

		# fifo file to control mplayer slave
		try:
			subprocess.check_output(["mkfifo","/tmp/mplayer-control"])
		except:
			pass

		# Set volume and output after init - might have changed while box was off ----
		bInitOutputSwitched   = False
		bInitVolumeAdjusted   = False

		while not(bInitOutputSwitched and bInitVolumeAdjusted):
			cCmd = readSerialMessageAndPlausibilize()
			if (cCmd[idxAction] == "Volume"):
				adjustVolume(cCmd[idxParam])
				bInitVolumeAdjusted = True
			elif (cCmd[idxAction] == "Output"):
				switchOutput(cCmd[idxParam])
				bInitOutputSwitched = True
			time.sleep(0.1)
			
		# look, if USB Stick is inserted
		try:
			subprocess.check_output(["ls","/dev/sda1"])
		except:
			time.sleep(15)
			Shutdown()
			pass		

		# mount USB Stick
		try:
			subprocess.check_output(["sudo","mkdir","/media/STICK"])
		except:
			pass
		
		try:
			subprocess.check_output(["sudo","mount","/dev/sda1","/media/STICK","-o","uid=pi,gid=pi,utf8,noatime"])
		except:
			pass
		time.sleep(numWaitAtInit)

		# if USB Stick was not prepared before -----------
		if not(lookIfFileExists(strBooksFolder + strUSBStickPrepared)):
			# read the booktitles in the books folders as 000.mp3 -----
			cFolders = listFiles(strBooksFolder,"Folders")
			for numBook in range(len(cFolders)):
				strBook = cFolders[numBook]
				readAsMP3(strBook,strBooksFolder + strBook + "/" + "000.mp3",strLanguagePattern)

				# read waitbar percent
				strPercent = str(int((float(numBook) / len(cFolders)) * 100))
				play(strText2SpeechFolder + str(strPercent).zfill(3),"end")
				play(strText2SpeechFolder + "Percent","end")
				print strPercent,"%"
			
			# put a marker, that it is ready
			subprocess.check_output(["touch",strBooksFolder + strUSBStickPrepared])

		strNextState = "Book"
		strParam =     "Actual"

	elif strState == "Book": # --------------------------------------------
		print "Book---"

		if strParam == "Next": # ------
			strBook = determineNextOrPreviousBook(strBooksFolder,strBook,strParam)
			setBookmark(strBook,strBooksFolder,strBookPrefix,strBookAppendix)

		elif strParam == "Previous": # ------
			strBook = determineNextOrPreviousBook(strBooksFolder,strBook,strParam)
			setBookmark(strBook,strBooksFolder,strBookPrefix,strBookAppendix)
			
		elif strParam == "Actual": # ------			
			# which book should be read first ? -----------
			cFolders = listFiles(strBooksFolder,"Folders")
			strDefaultBook = cFolders[0]
			strBook = getBookmark(strDefaultBook,strBooksFolder,strBookPrefix,strBookAppendix)

		# stop any mp3 playback
		play("","stop")

		# read book title ----
		play(strBooksFolder + strBook + "/" + "000","end")

		# check if "next book" is still pressed
		cCmd = readSerialMessageAndPlausibilize()
		if (cCmd[idxAction]=="Book"):
			strNextState = "Book"
			strParam     = cCmd[idxParam]
		else:
			strNextState = "Chapter"
			strParam =     "Actual"

	elif strState == "Chapter": # --------------------------------------------
		print "Chapter---"

		if strParam == "Actual": # ------
			strFileNum = getBookmark("001",strBooksFolder + strBook + "/",strChapterPrefix,strChapterAppendix)

		elif strParam == "Next": # ------
			# look, if filename + 1 exists
			strFileNum = str(int(strFileNum) + 1).zfill(3)
			bExists = lookIfFileExists(strBooksFolder + strBook + "/" +  strFileNum + ".mp3")
			if not bExists:
				strFileNum = "001"
			setBookmark(strFileNum, strBooksFolder + strBook + "/",strChapterPrefix,strChapterAppendix)

		elif strParam == "Previous": # ------
			strFileNum = str(int(strFileNum) - 1).zfill(3)
			if strFileNum == "000":
				strFileNum = findLastMP3File(strBooksFolder + strBook)
			setBookmark(strFileNum, strBooksFolder + strBook + "/",strChapterPrefix,strChapterAppendix)

		# stop any mp3 playback
		play("","stop")

		# say e.g. "42" for chapter 42
		play(strText2SpeechFolder + "num/" + strFileNum, "end")

		# check if "next chapter" is still pressed
		cCmd = readSerialMessageAndPlausibilize()
		if (cCmd[idxAction]=="Chapter"):
			strNextState = "Chapter"
			strParam     = cCmd[idxParam]
		else:
			# play the audio book
			play(strBooksFolder + strBook + "/" + strFileNum, "non-blocking")

			strNextState = "Play"
			strParam =     ""

	elif strState == "Play": # --------------------------------------------
		if strLastState == "Pause":
			print "Play---"
			numTimeCount = 0

		elif strLastState == "Play":
			cCmd = readSerialMessageAndPlausibilize()
			strAction = cCmd[idxAction]
			strParam = cCmd[idxParam]

			# a button is pressed
			if len(strAction)>0:
				if strAction == "Pause":
					if numTimeCount >= 0:
						numTimeCount = numTimeCount + 1
						if numTimeCount > 20:
							strNextState = "Shutdown"
					else:
						strNextState = "Pause"
				else:
					strNextState = strAction

			else:  # so, there is no button pressed: at least no shutdown wanted
				numTimeCount = -1
				
				# check status - if ended after playing then play next chapter
				bStillRunning = False
				try:
					strShell = subprocess.check_output(["pidof","mplayer"])
					bStillRunning = True
				except:
					pass
				
				if bStillRunning == False:
					strNextState = "Chapter"
					strParam  = "Next"

	elif strState == "Volume": # --------------------------------------------
		print "Volume---"
		adjustVolume(strParam)
		strNextState = strLastState

	elif strState == "Output": # --------------------------------------------
		print "Output---"
		switchOutput(strParam)
		strNextState = strLastState

	elif strState == "Pause": # --------------------------------------------
		if strLastState == "Play":
			print "Pause---"
			numTimeCount = 0
			play("","pause")

		elif strLastState == "Pause":
			cCmd = readSerialMessageAndPlausibilize()
			strAction = cCmd[idxAction]
			strParam = cCmd[idxParam]

			if len(strAction)>0:
				if strAction == "Pause":
					if numTimeCount >= 0:
						numTimeCount = numTimeCount + 1
						if numTimeCount > 20:
							strNextState = "Shutdown"
					else:
						print "Unpause..."
						strNextState = "Play"
						play("","unpause")

				elif strAction == "Volume":
					strNextState = "Volume"
				elif strAction == "Output":
					strNextState = "Output"
			else:  # so, there is no button pressed: no shutdown wanted
				numTimeCount = -1

	elif strState == "Shutdown": # --------------------------------------------
		print "Shutdown---"
		Shutdown()

	strLastState = strState
	strState = strNextState
	time.sleep(0.1)

